#include "nmSynth.hpp"
#include "nmSynthEdit.hpp"
#include "nmMidiControl.hpp"
#include "nmSynthControl.hpp"
#include <windows.h>

class nmSynthProgram
{
  friend class nmSynth;

  float wave_type;
  float attack_time;
  float attack_level;
  float decay_time;
  float sustain_level;
  float release_time;
  float low_pass;
  float resonance;
  char name[24+1];
public:
  enum{
    WAVE_TYPE = 0,
    ATTACK_TIME,
    ATTACK_LEVEL,
    DECAY_TIME,
    SUSTAIN_LEVEL,
    RELEASE_TIME,
    LOW_PASS,
    RESONANCE,
    NUM_PARAMS
  };
  enum{
    NUM_PROGS = 4
  };
  nmSynthProgram();
  ~nmSynthProgram(){}
};

nmSynthProgram::nmSynthProgram(){
  wave_type = 0.5f;
  attack_time = 0.0f;
  attack_level = 0.3f;
  decay_time = 0.0f;
  sustain_level = 0.3f;
  release_time = 0.0f;
  low_pass = 0.999f;
  resonance = 0.999f;
}

nmSynth::nmSynth(audioMasterCallback audioMaster)
: AudioEffectX(audioMaster, nmSynthProgram::NUM_PROGS, nmSynthProgram::NUM_PARAMS), sc(NULL)
{
  int i;

  programs = new nmSynthProgram[nmSynthProgram::NUM_PROGS];

  for(i = 0; i < nmSynthProgram::NUM_PROGS; i++){
    sprintf(programs[i].name, "Default %d", i);
  }

  try{
    vwe = new nmSynthEdit(this);
  }
  catch(int){
    delete [] programs;
    throw -1;
  }

  mc = new nmMidiControl();
  sc = new nmSynthControl();

  setEditor(vwe);

  setWaveType(0.5f);
  setLowPass(0.999f, 0.999f);
  setEnvelope(0.0f, 0.3f, 0.0f, 0.3f, 0.0f);

  if(programs){
    setProgram(0);
  }

  setNumInputs(0);                  // stereo in
  setNumOutputs(2);                 // stereo out
  setUniqueID('nmS ');
  isSynth(true);                    // synth on
  canProcessReplacing();            // supports both accumulating and replacing output

  events = NULL;
  eventMaxNum = 0;
  eventNum = 0;
  currentEventNo = 0;
  currentSample = 0;

  suspend();
}

void nmSynth::setWaveType(const float &wt){
  nmSynthProgram *ap = &programs[curProgram];
  ap->wave_type = wt;
  if(wt < 0.33){
    sc->setWaveType(0);
  }
  else if(wt > 0.66){
    sc->setWaveType(2);
  }
  else{
    sc->setWaveType(1);
  }
}

void nmSynth::setLowPass(const float &lp, const float &res){
  nmSynthProgram *ap = &programs[curProgram];
  ap->low_pass = lp;
  ap->resonance = res;
  sc->setLowPass(lp, res);
}

void nmSynth::setEnvelope(const float &at, const float &al, const float &dt, const float &sl, const float &rt){
  nmSynthProgram *ap = &programs[curProgram];
  ap->attack_time = at;  
  ap->attack_level = al;
  ap->decay_time = dt;
  ap->sustain_level = sl;
  ap->release_time = rt;
  sc->setEnvelope(at, al, dt, sl, rt);
}

nmSynth::~nmSynth(){
  delete sc;
  delete mc;
  delete vwe;
  vwe = NULL;
  setEditor(NULL);
  if(programs){
    delete [] programs;
  }
}

void nmSynth::setProgram (long program){
  if(program < 0 || program >= nmSynthProgram::NUM_PROGS){
    return;
  }

  nmSynthProgram *ap = &programs[program];
  curProgram = program;
  
  setParameter(nmSynthProgram::WAVE_TYPE, ap->wave_type);
  setParameter(nmSynthProgram::ATTACK_TIME, ap->attack_time);
  setParameter(nmSynthProgram::ATTACK_LEVEL, ap->attack_level);
  setParameter(nmSynthProgram::DECAY_TIME, ap->decay_time);
  setParameter(nmSynthProgram::SUSTAIN_LEVEL, ap->sustain_level);
  setParameter(nmSynthProgram::RELEASE_TIME, ap->release_time);
  setParameter(nmSynthProgram::LOW_PASS, ap->low_pass);
  setParameter(nmSynthProgram::RESONANCE, ap->resonance);

  if(sc != NULL){
    setWaveType(ap->wave_type);
    setLowPass(ap->low_pass, ap->resonance);
    setEnvelope(ap->attack_time, ap->attack_level, ap->decay_time, ap->sustain_level, ap->release_time);
  }

  if(vwe != NULL){
    vwe->setDial(ap->wave_type, ap->low_pass, ap->resonance, ap->attack_time, ap->attack_level, ap->decay_time, ap->sustain_level, ap->release_time);
  }
}

void nmSynth::setProgramName(char *name){
  strncpy(programs[curProgram].name, name, 24);
}

void nmSynth::getProgramName(char *name){
  strncpy(name, programs[curProgram].name, 24);
}

void nmSynth::setParameter (long index, float value){
  nmSynthProgram *ap = &programs[curProgram];
  switch(index){
    case nmSynthProgram::WAVE_TYPE:
      ap->wave_type  = value;
      break;
    case nmSynthProgram::ATTACK_TIME:
      ap->attack_time = value;
      break;
    case nmSynthProgram::ATTACK_LEVEL:
      ap->attack_level = value;
      break;
    case nmSynthProgram::DECAY_TIME:
      ap->decay_time = value;
      break;
    case nmSynthProgram::SUSTAIN_LEVEL:
      ap->sustain_level = value;
      break;
    case nmSynthProgram::RELEASE_TIME:
      ap->release_time = value;
      break;
    case nmSynthProgram::LOW_PASS:
      ap->low_pass = value;
      break;
    case nmSynthProgram::RESONANCE:
      ap->resonance = value;
      break;
    default:
      break;
  }
}

float nmSynth::getParameter(long index){
  float value = 0;
  nmSynthProgram *ap = &programs[curProgram];
  switch(index){
    case nmSynthProgram::WAVE_TYPE:
      value = ap->wave_type;
      break;
    case nmSynthProgram::ATTACK_TIME:
      value = ap->attack_time;
      break;
    case nmSynthProgram::ATTACK_LEVEL:
      value = ap->attack_level;
      break;
    case nmSynthProgram::DECAY_TIME:
      value = ap->decay_time;
      break;
    case nmSynthProgram::SUSTAIN_LEVEL:
      value = ap->sustain_level;
      break;
    case nmSynthProgram::RELEASE_TIME:
      value = ap->release_time;
      break;
    case nmSynthProgram::LOW_PASS:
      value = ap->low_pass;
      break;
    case nmSynthProgram::RESONANCE:
      value = ap->resonance;
      break;
    default:
      break;
  }

  return value;
}

bool nmSynth::getOutputProperties(long index, VstPinProperties* properties){
  if(index < 2){
    sprintf(properties->label, "nmS %02d", index + 1);
    properties->flags = kVstPinIsActive;
    if(index < 2){
      properties->flags |= kVstPinIsStereo;  // test, make channel 1+2 stereo
    }
    return true;
  }
  return false;
}

long nmSynth::canDo (char* text){
  if(!strcmp (text, "receiveVstEvents"))
    return 1;
  if(!strcmp (text, "receiveVstMidiEvent"))
    return 1;
  if(!strcmp (text, "midiProgramNames"))
    return 1;
  return -1;  // explicitly can't do; 0 => don't know
}

// Tv[g̐ݒ
void nmSynth::setSampleRate(float sampleRate){
  AudioEffectX::setSampleRate(sampleRate);
  sc->setSampleRate(sampleRate);
}

// GtFNgĩvÕCj
void nmSynth::process(float **inputs, float **outputs, long sampleFrames){
  float *out1 = outputs[0];
  float *out2 = outputs[1];

  for(int i = 0; i < sampleFrames; i++){
    (*out1++) = 0.0f;
    (*out2++) = 0.0f;
  }
}

#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

// GtFNgĩvÕCj
void nmSynth::processReplacing(float **inputs, float **outputs, long sampleFrames){
  float *out1 = outputs[0];
  float *out2 = outputs[1];

  nRestSample = sampleFrames;

  while(nRestSample > 0){
    if(currentEventNo < eventNum){
      pE = (VstMidiEvent*)&events[currentEventNo];
      nProcessSample = pE->deltaFrames - currentSample;
      nProcessSample = MAX(0, MIN(nRestSample, nProcessSample));
    }else{
      pE = NULL;
      nProcessSample = nRestSample;
    }

    nRestSample2 = nProcessSample;
    while(nRestSample2 > 0){
      if(sc->isSilence() != true){
        float f = (float)sc->getWave();

        if(f > 32767.0f) f =  32767.0f;
        if(f < -32768.0f) f = -32768.0f;

        (*out1++) = f * mc->refrectMidiLeftState();
        (*out2++) = f * mc->refrectMidiRightState();
      }
      else{
        (*out1++) = 0.0f;
        (*out2++) = 0.0f;
      }

      nRestSample2 --;
    }

    currentSample += nProcessSample;

    if(pE){
      long first = pE->midiData[0] & 0xf0;
      long second = pE->midiData[1];
      long third = pE->midiData[2];

      switch(first){  
        case 0x80:
          sc->offNote(second);
          break;
        case 0x90:
          sc->setPitch(64);
          sc->setVibrato(0);
          sc->setNote(second, third);
          break;
        case 0xB0:
          switch(second){
            case 0x01:
              sc->setVibrato(third);
              break;
            case 0x0A:
              mc->setPan(third);
              break;
            case 0x07:
              mc->setVolume(third);
              break;
            case 0x7e:
            case 0x7b:
              sc->offAllNote();
              break;
            default:
              break;
          }
          break;
        case 0xE0:
          sc->setPitch(third);
          break;
        default:
          break;
      }

      currentEventNo++;
    }

    nRestSample -= nProcessSample;
  }
}

void nmSynth::resume(){
  wantEvents();
}

// Cxg󂯎蕔
long nmSynth::processEvents(VstEvents* ev){
  currentEventNo = 0;
  currentSample = 0;
  eventNum = 0;
  if(ev->numEvents > 0){
    if(eventMaxNum < ev->numEvents){
      if(events){
        delete [] events;
        events = NULL;
      }
      eventMaxNum = ev->numEvents;
      events = new VstMidiEvent[eventMaxNum];
    }
    for(int i = 0; i < ev->numEvents; i++){
      if(ev->events[i]->type == kVstMidiType){
        events[eventNum] = *(VstMidiEvent*)(ev->events[i]);
        eventNum++;
      }
    }
  }

  return 1;
}
